#include "plugin.h"

#include <sdk_util.h>

/* Grapple Web 
 * -
 * Based off of the tutorial at http://www.planethalflife.com/hlprogramming/old/tuts/tut_offhandgrapple.htm
 * By "EvilClone"
 * -
 * Allow gorges to emit a grappling hook-like web strand to pull them to their destination.
 * Cvars:
 *   mm_grapple_hives  (2)    - How many hives this feature requires.
 *   mm_grapple_speed  (500)  - Speed the gorge goes up the hook at.
 *   mm_grapple_hspeed (1000) - Speed the hook flies at.
 *   mm_grapple_range  (1000) - Range a grapple can reach.
 *   mm_grapple_delay  (0.5)  - Seconds it takes from the hook hitting a target before it starts pulling.
 *   mm_grapple_combat (1)    - Toggles grapple web feature on or off for combat mode.
 * Commands:
 *   +hook - Initiate the grapple.
 *   -hook - Stop the grapple.
 *
 * NOTE: You must change the map for cvar changes to take effect!
 */

#define WEB_RED			64
#define WEB_GREEN		64
#define WEB_BLUE		64
#define WEB_BRIGHT		128
#define MASK_SILENCE	256

float fNextHook[33];
int gmsgHudText2=0;
int hookedmsg=0;

int iCvarsRead=0;
int iWebIndex;
int iStrandIndex;
unsigned short iEventIndex=1;

int iCombat=0;

int HookHookSpeed;
int HookSpeed;
float HookRange;
int HookHives;
float HookDelay;
int HookCombat;

bool AtEnd[33];
bool InHook[33];
edict_t *PtoHook[33];
Vector HangingLocation[33];

#define HOOK_FIRING		1
#define HOOK_SNAGGED	2
#define HOOK_PULLING	3
#define HOOK_FINISHED	4

void MakeHook(edict_t *pOwner);
bool IsTargetSolid(edict_t *pEntity);
bool IsValidHookTarget(edict_t *pEntity);
void KillHook( edict_t *pHook );
edict_t *UTIL_FindEntityByString(edict_t *pentStart, const char *szKeyword, const char *szValue);
int GetHiveCount(void);
bool CanIHook(edict_t *pEntity);
void KillAllHooks(void);
void ReadCvars(void);
bool IsCombat(void);

int Spawn(edict_t *pEntity) {
	int i;
	if(FStrEq(STRING(pEntity->v.classname), "worldspawn"))
	{
		/* Check if we're in combat or not */
		iCombat=0;
		/*
		char mapname[255];
		strcpy(mapname,STRING(gpGlobals->mapname));
		if (mapname[0]=='c' && mapname[1]=='o' && mapname[2]=='_')
			iCombat=1;
		*/
		iCvarsRead=0;
		// New round starting.
		/* Precache the sprites for the grapple web. */
		for (i=0;i<=32;i++)
		{
			AtEnd[i]=false;
			InHook[i]=false;
			PtoHook[i]=NULL;
			fNextHook[i]=0.0;
		}
		gmsgHudText2=0;
		iStrandIndex=PRECACHE_MODEL("sprites/webstrand.spr");
		iWebIndex=PRECACHE_MODEL("sprites/webprojectile.spr");
		PRECACHE_SOUND("weapons/webspin1.wav");
		PRECACHE_SOUND("weapons/webspin2.wav");
		//iEventIndex=PRECACHE_EVENT(1,"events/Particle.sc");
		iEventIndex=52;
		RETURN_META_VALUE(MRES_HANDLED, 0);
	}

	RETURN_META_VALUE(MRES_IGNORED, 0);
}
void ClientCommand( edict_t *pEntity )
{
	const char* szUserCommand = CMD_ARGV(0);
	if( szUserCommand[0] == '\0' ) return;
	if (FStrEq(CMD_ARGV(0),"jointeamone") || FStrEq(CMD_ARGV(0),"jointeamtwo") || FStrEq(CMD_ARGV(0),"autoassign") || FStrEq(CMD_ARGV(0),"spectate") || FStrEq(CMD_ARGV(0),"readyroom"))
	{
		if (InHook[ENTINDEX(pEntity)])
		{
			KillHook(PtoHook[ENTINDEX(pEntity)]);
			InHook[ENTINDEX(pEntity)]=0;
		}
		if (pEntity->v.team == 0)
		{
			fNextHook[ENTINDEX(pEntity)]=gpGlobals->time;
		}
	}
	if (FStrEq(CMD_ARGV(0),"+hook"))
	{
		if (iCvarsRead==0)
		{
			iCvarsRead=1;
			iCombat=IsCombat();
			ReadCvars();
		}
		if (CanIHook(pEntity))
		{
			MakeHook(pEntity);
		}
		RETURN_META(MRES_SUPERCEDE);
	}
	if (FStrEq(CMD_ARGV(0),"-hook"))
	{
		if (InHook[ENTINDEX(pEntity)])
		{
			KillHook(PtoHook[ENTINDEX(pEntity)]);
		}
		InHook[ENTINDEX(pEntity)]=false;
		RETURN_META(MRES_SUPERCEDE);
	}
	RETURN_META(MRES_IGNORED);
}
void MakeHook( edict_t *pOwner )
{
	// Prevent double hooking
	if (InHook[ENTINDEX(pOwner)])
		return;
	InHook[ENTINDEX(pOwner)]=true;
	edict_t *pEntity = CREATE_NAMED_ENTITY(MAKE_STRING("info_target"));
	pEntity->v.classname=MAKE_STRING("web_grapple");
	pEntity->v.movetype=MOVETYPE_FLY;
	pEntity->v.solid=SOLID_BBOX;
	pEntity->v.owner=pOwner;
	pEntity->v.gravity=0.5;
	SET_MODEL(pEntity,"sprites/webprojectile.spr");
	SET_ORIGIN(pEntity,pOwner->v.origin);
	SET_SIZE(pEntity,Vector(-1,-1,-1),Vector(1,1,1));
	pEntity->v.iuser1=HOOK_FIRING;
	pEntity->v.nextthink=gpGlobals->time + 0.1;
	pEntity->v.framerate=10.0;
	PtoHook[ENTINDEX(pOwner)]=pEntity;
	// Find the way the player is facing...
	Vector v_aim;
	Vector v_up;
	g_engfuncs.pfnAngleVectors(pOwner->v.v_angle + pOwner->v.punchangle,v_aim,v_up,NULL);
	//v_aim.x = -v_aim.x;
	pEntity->v.origin = (pOwner->v.origin + pOwner->v.view_ofs);// - (v_up * 2);
	pEntity->v.velocity = v_aim * HookHookSpeed;
	pEntity->v.speed = HookHookSpeed;
	if (!(pOwner->v.iuser4 & MASK_SILENCE)) // Don't let out the sound if we have silence...
		EMIT_SOUND(pOwner, CHAN_ITEM, "weapons/webspin1.wav", VOL_NORM, ATTN_NORM);
	/*
	MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
    WRITE_BYTE( TE_BEAMENTS ); 
    WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
    WRITE_SHORT( ENTINDEX(pEntity) );
    WRITE_SHORT( iStrandIndex ); 
    WRITE_BYTE( 1 );
    WRITE_BYTE( 0 ); 
    WRITE_BYTE( 1 ); 
    WRITE_BYTE( 10 ); 
    WRITE_BYTE( 0 ); 
    WRITE_BYTE( 128 ); //Red
    WRITE_BYTE( 128 ); //Green
    WRITE_BYTE( 128 ); //Blue
    WRITE_BYTE( 128 ); //Brightness
    WRITE_BYTE( 0);
	MESSAGE_END( );
	*/

}
void KillHook( edict_t *pHook )
{
	InHook[ENTINDEX(pHook->v.owner)]=false;
    pHook->v.owner->v.movetype = MOVETYPE_WALK;
    pHook->v.owner->v.gravity = 1;
	AtEnd[ENTINDEX(pHook->v.owner)]=false;
	InHook[ENTINDEX(pHook->v.owner)]=false;
	PtoHook[ENTINDEX(pHook->v.owner)]=NULL;
	HangingLocation[ENTINDEX(pHook->v.owner)]=Vector(0,0,0);
	pHook->v.flags |= FL_KILLME;
	pHook->v.iuser1 = HOOK_FINISHED;

	pHook->v.nextthink = gpGlobals->time;
}
void Think( edict_t *pEntity )
{
	if (FStrEq(STRING(pEntity->v.classname),"web_grapple"))
	{
		if (!InHook[ENTINDEX(pEntity->v.owner)] || !CanIHook(pEntity->v.owner))
		{
			KillHook(pEntity);
		}
		else
		{
			if (pEntity->v.iuser1==HOOK_FIRING)
			{
				if (pEntity->v.frame == 3)
				{
					pEntity->v.frame=0;
				}
				else
					pEntity->v.frame += 1;
				MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
				WRITE_BYTE( TE_BEAMENTS ); 
				WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
				WRITE_SHORT( ENTINDEX(pEntity) );
				WRITE_SHORT( iStrandIndex ); 
				WRITE_BYTE( 1 );
				WRITE_BYTE( 0 ); 
				WRITE_BYTE( 1 ); 
				WRITE_BYTE( 10 ); 
				WRITE_BYTE( 0 ); 
				WRITE_BYTE( WEB_RED ); //Red
				WRITE_BYTE( WEB_GREEN ); //Green
				WRITE_BYTE( WEB_BLUE ); //Blue
				WRITE_BYTE( WEB_BRIGHT ); //Brightness
				WRITE_BYTE( 0);
				MESSAGE_END( );
	
			}
			else if (pEntity->v.iuser1==HOOK_SNAGGED)
			{
				MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
				WRITE_BYTE( TE_BEAMENTS ); 
				WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
				WRITE_SHORT( ENTINDEX(pEntity) );
				WRITE_SHORT( iStrandIndex ); 
				WRITE_BYTE( 1 );
				WRITE_BYTE( 0 ); 
				WRITE_BYTE( 1 ); 
				WRITE_BYTE( 10 ); 
				WRITE_BYTE( 0 ); 
				WRITE_BYTE( WEB_RED ); //Red
				WRITE_BYTE( WEB_GREEN ); //Green
				WRITE_BYTE( WEB_BLUE ); //Blue
				WRITE_BYTE( WEB_BRIGHT ); //Brightness
				WRITE_BYTE( 0);
				MESSAGE_END( );
				if (!(AtEnd[ENTINDEX(pEntity->v.owner)]))
				{
					if ((pEntity->v.owner->v.origin - pEntity->v.origin ).Length() >= HookRange)
					{
						KillHook(pEntity);
						RETURN_META(MRES_IGNORED);
					}
					if((pEntity->v.owner->v.origin - pEntity->v.origin ).Length() >= 50.0)
					{
				        pEntity->v.owner->v.velocity = (pEntity->v.origin - pEntity->v.owner->v.origin) * 10; 
					    pEntity->v.owner->v.speed = HookSpeed;
			        } 
					else 
					{
						HangingLocation[ENTINDEX(pEntity->v.owner)]=pEntity->v.owner->v.origin;
					    pEntity->v.owner->v.origin = HangingLocation[ENTINDEX(pEntity->v.owner)];
				        pEntity->v.owner->v.velocity = Vector(0, 0, 0);
					    pEntity->v.owner->v.gravity = 0;
						pEntity->v.owner->v.speed = 0;
						AtEnd[ENTINDEX(pEntity->v.owner)] = true;
					}
				}
				else
				{
					if ((pEntity->v.owner->v.origin - pEntity->v.origin ).Length() >= 300)
					{
						KillHook(pEntity);
						RETURN_META(MRES_IGNORED);
					}
					pEntity->v.owner->v.origin = HangingLocation[ENTINDEX(pEntity->v.owner)];
			        pEntity->v.owner->v.velocity = Vector(0, 0, 0);
				    pEntity->v.owner->v.gravity = 0;
					pEntity->v.owner->v.speed = 0;
				}
			}
			else if (pEntity->v.iuser1 == HOOK_PULLING)
			{
			}
			pEntity->v.nextthink=gpGlobals->time + 0.1;
		}
	}
	RETURN_META(MRES_IGNORED);
}
void Touch(edict_t *pentTouched, edict_t *pEntity)
{
	if (FStrEq(STRING(pEntity->v.classname),"player"))
	{
		if (FStrEq(STRING(pentTouched->v.classname),"info_join_team") || FStrEq(STRING(pentTouched->v.classname),"info_spectate") || FStrEq(STRING(pentTouched->v.classname),"info_join_autoassign"))
		{
			if (InHook[ENTINDEX(pEntity)])
			{
				KillHook(PtoHook[ENTINDEX(pEntity)]);
				InHook[ENTINDEX(pEntity)]=0;
			}
			if (pEntity->v.team == 0)
			{
				fNextHook[ENTINDEX(pEntity)]=gpGlobals->time;
			}
			RETURN_META(MRES_IGNORED);
		}
	}
	if (FStrEq(STRING(pEntity->v.classname),"web_grapple"))
	{
		if (pentTouched != pEntity->v.owner)
		{
			if (!IsValidHookTarget(pentTouched))
			{
				if (!IsTargetSolid(pentTouched))
				{
					RETURN_META(MRES_IGNORED);
				}
				else
				{
					MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
					WRITE_BYTE( TE_BEAMENTS ); 
					WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
					WRITE_SHORT( ENTINDEX(pEntity) );
					WRITE_SHORT( 0 ); 
					WRITE_BYTE( 0 );
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); //Red
					WRITE_BYTE( 0 ); //Green
					WRITE_BYTE( 0 ); //Blue
					WRITE_BYTE( 0 ); //Brightness
					WRITE_BYTE( 0);
					MESSAGE_END( );
					MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
					WRITE_BYTE( TE_BEAMENTS ); 
					WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
					WRITE_SHORT( ENTINDEX(pEntity) );
					WRITE_SHORT( 0 ); 
					WRITE_BYTE( 0 );
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); 
					WRITE_BYTE( 0 ); //Red
					WRITE_BYTE( 0 ); //Green
					WRITE_BYTE( 0 ); //Blue
					WRITE_BYTE( 0 ); //Brightness
					WRITE_BYTE( 0);
					MESSAGE_END( );
					KillHook(pEntity);
					RETURN_META(MRES_SUPERCEDE);
				}
			}
			if (!IsTargetSolid(pentTouched))
			{
				RETURN_META(MRES_IGNORED);
			}
			if (!(pEntity->v.owner->v.iuser4 & MASK_SILENCE)) // Don't let out the sound if we have silence...
				EMIT_SOUND(pEntity, CHAN_ITEM, "weapons/webspin2.wav", VOL_NORM, ATTN_NORM);
			pEntity->v.effects |= EF_NODRAW;
			PLAYBACK_EVENT_FULL(0,pEntity,iEventIndex,0.0,pEntity->v.origin,Vector(0,0,0),0.0,0.0,9,0,0,0);
			pEntity->v.iuser1 = HOOK_SNAGGED;
			pEntity->v.velocity=Vector(0,0,0);
			pEntity->v.solid=SOLID_NOT;
			pEntity->v.movetype=MOVETYPE_FLY;
			pEntity->v.nextthink = gpGlobals->time + HookDelay;;
			MESSAGE_BEGIN( MSG_BROADCAST, SVC_TEMPENTITY ); //Draw the rope
			WRITE_BYTE( TE_BEAMENTS ); 
			WRITE_SHORT( ENTINDEX(pEntity->v.owner) );
			WRITE_SHORT( ENTINDEX(pEntity) );
			WRITE_SHORT( iStrandIndex ); 
			WRITE_BYTE( 1 );
			WRITE_BYTE( 0 ); 
			WRITE_BYTE( (int)(HookDelay * 10) + 1 ); 
			WRITE_BYTE( 10 ); 
			WRITE_BYTE( 0 ); 
			WRITE_BYTE( WEB_RED ); //Red
			WRITE_BYTE( WEB_GREEN ); //Green
			WRITE_BYTE( WEB_BLUE ); //Blue
			WRITE_BYTE( WEB_BRIGHT ); //Brightness
			WRITE_BYTE( 0);
			MESSAGE_END( );


		}
	}
	RETURN_META(MRES_IGNORED);
}
/*
  Addition to the valid target detection
*/
bool IsTargetSolid(edict_t *pEntity)
{
	if (pEntity->v.solid == SOLID_NOT)
		return false;
	if (pEntity->v.solid == SOLID_TRIGGER)
		return false;
	return true;
}
/*
   List of solid targets I can not hook.
*/
bool IsValidHookTarget(edict_t *pEntity)
{
	if (FStrEq(STRING(pEntity->v.classname),"defensechamber"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"offensechamber"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"movementchamber"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"sensorychamber"))
		return false;
/*	if (FStrEq(STRING(pEntity->v.classname),"team_hive"))
		return false;
*/
	if (FStrEq(STRING(pEntity->v.classname),"player"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_command"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"alienresourcetower"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"resourcetower"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"turret"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"siegeturret"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"phasegate"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_turretfactory"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_advturretfactory"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_armory"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_advarmory"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_armslab"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_protolab"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_observatory"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"team_infportal"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"info_join_team"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"info_spectate"))
		return false;
	if (FStrEq(STRING(pEntity->v.classname),"info_join_autoassign"))
		return false;
	return true;
}
void PlayerPreThink(edict_t *pEntity)
{
	if (pEntity->v.impulse == 5 && InHook[ENTINDEX(pEntity)])
	{
		KillHook(PtoHook[ENTINDEX(pEntity)]);
		InHook[ENTINDEX(pEntity)]=0;
	}
	RETURN_META(MRES_IGNORED);
}
edict_t *UTIL_FindEntityByString(edict_t *pentStart, const char *szKeyword, const char *szValue)
{
	edict_t *pentEntity;
	pentEntity=FIND_ENTITY_BY_STRING(pentStart, szKeyword, szValue);
	if(!FNullEnt(pentEntity))
		return pentEntity;
	return NULL;
}
/*
   # of active hives.
*/
int GetHiveCount(void)
{
	int iCount=0;
	edict_t *pEntity=NULL;
	while ((pEntity = UTIL_FindEntityByString(pEntity,"classname","team_hive")) != NULL)
	{
		if (pEntity->v.fuser1 >= 1000)
			iCount++;
	}
	return iCount;
}
/*
   Whether or not I can shoot a hook.
*/
bool CanIHook(edict_t *pEntity)
{
	if (gpGlobals->time < fNextHook[ENTINDEX(pEntity)])
		return false;
	if (pEntity->v.deadflag > 0)
		return false;
	if (FStrEq(STRING(pEntity->v.model),"models/player/alien2/alien2.mdl") && GetHiveCount() >= HookHives)
		return true;
	if (FStrEq(STRING(pEntity->v.model),"models/player/alien2/alien2.mdl") && pEntity->v.team == 0) // Ready room
		return true;
	if (FStrEq(STRING(pEntity->v.model),"models/player/alien2/alien2.mdl") && HookCombat == 1 && iCombat == 1)
		return true;
	return false;
	
}
/*
   Properly remove all hooks from the game.
*/
void KillAllHooks(void)
{
	int i=0;
	for (i=0;i<=32;i++)
	{
		if (InHook[i])
		{
			KillHook(PtoHook[i]);
		}
	}
}

/* 
   Read cvars when the first +hook command is issued.
*/
void ReadCvars(void)
{
	HookHookSpeed=(int)cvar_mm_grapple_hspeed->value;
	HookSpeed=(int)cvar_mm_grapple_speed->value;
	HookHives=(int)cvar_mm_grapple_hives->value;
	HookRange=cvar_mm_grapple_range->value;
	HookDelay=cvar_mm_grapple_delay->value;
	HookCombat=(int)cvar_mm_grapple_combat->value;
	if (HookDelay > 25.5) // Or it crashes... 
		HookDelay = 25.5;
	if (HookDelay < 0.0)
		HookDelay = 0.0;
}

/* 
   Return if the map is combat or not, depending on number of hives.
   My old, surefire way of getting combat is getting reports of crashes (???)
   So this is a quick replacement.
*/
bool IsCombat(void)
{
	int iCount=0;
	edict_t *pEntity=NULL;
	while ((pEntity = UTIL_FindEntityByString(pEntity,"classname","team_hive")) != NULL)
	{
			iCount++;
	}
	if (iCount==1)
		return true;
	return false;
}

void MessageBegin(int msg_dest, int msg_type, const float *pOrigin, edict_t *pEntity)
{
	if (gmsgHudText2 == 0)
		gmsgHudText2 = GET_USER_MSG_ID(&Plugin_info,"HudText2",NULL);
	hookedmsg=0;
	if (msg_type == gmsgHudText2)
	{
		hookedmsg=gmsgHudText2;
	}
	RETURN_META(MRES_IGNORED);
}
void MessageEnd(void)
{
	hookedmsg=0;
	RETURN_META(MRES_IGNORED);
}
void WriteString(const char *sz)
{
	if (hookedmsg==gmsgHudText2)
	{
		if (FStrEq("TeamTwoWon",sz) || FStrEq("TeamOneWon",sz) || FStrEq("GameDraw",sz))
		{
			hookedmsg=0;
			KillAllHooks();
			int k;
			for (k=0;k<=32;k++)
			{
				fNextHook[k]=gpGlobals->time + 15.0;
			}
		}
	}
	RETURN_META(MRES_IGNORED);
}
